/* perform tests according to the SHAVS specification */

#include <stdio.h>
#include <string.h>
#include <stdlib.h>
#include <ctype.h>
#include <errno.h>

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "../src/hash.h"
#include "../src/sha1.h"
#include "../src/sha256.h"
#include "../src/crc32.h"

#define MAXLINE 200  /* maximal line length */

/* ---------------------------------------------------------------------- */
/* auxiliary functions */

/* write bytes to file in hexadecimal */
static inline void fprint_hex(FILE *fout, unsigned char *p, int n) {
  int i;
  
  for (i=0; i<n; i++) {
    fprintf(fout, "%02x", p[i]);
  }
}

/* convert a hexadecimal digit to a number. Return 0-15 on success,
   -1 if invalid digits, -2 if '\0' or EOF. */
static inline int digit_to_hex(int c) {
  if (c == 0 || c == EOF) {
    return -2;
  }
  if (!isxdigit(c)) {
    return -1;
  }
  return ((c >= 'A') ? c-'A'+10 : c) & 0xf;
}

/* read a hexadecimal digit from file, possibly after white space.
   Return 0-15 on success, -1 if invalid digits, -2 if end of file. */
static inline int read_hexdigit(FILE *fin) {
  int c;
  c = fgetc(fin);
  while (c == ' ' || c == '\t' || c == '\n' || c == '\r') {
    c = fgetc(fin);
  }
  return digit_to_hex(c);
}

/* read n hexadecimal bytes from file, possibly containing white
   space. Return 0 on success, -1 if invalid digits, -2 on end of
   file. */
static inline int read_hex(FILE *fin, unsigned char *p, int n) {
  int i, hi, lo;

  for (i=0; i<n; i++) {
    hi = read_hexdigit(fin);
    if (hi<0) {
      return hi;
    }
    lo = read_hexdigit(fin);
    if (lo<0) {
      return lo;
    }
    p[i] = (hi << 4) | lo;
  }
  return 0;
}

/* read n hexadecimal bytes from string, possibly containing white
   space. Return 0 on success, -1 if invalid digits, -2 on end of
   string. */
static inline int string_hex(char *str, unsigned char *p, int n) {
  int i, hi, lo;

  for (i=0; i<n; i++) {
    while (*str == ' ' || *str == '\t' || *str == '\n' || *str == '\r') {
      str++;
    }
    hi = digit_to_hex(*str);
    str++;
    if (hi<0) {
      return hi;
    }
    while (*str == ' ' || *str == '\t' || *str == '\n' || *str == '\r') {
      str++;
    }
    lo = digit_to_hex(*str);
    str++;
    if (lo<0) {
      return lo;
    }
    p[i] = (hi << 4) | lo;
  }
  return 0;
}

/* skip to next end-of-line. Return 0 on success, -1 on EOF */
static inline int skip_line(FILE *fin) {
  int c;

  c = fgetc(fin);
  while (c != '\n' && c != EOF) {
    c = fgetc(fin);
  }
  if (c == EOF) {
    return -1;
  } 
  return 0;
}

/* ---------------------------------------------------------------------- */

/* perform ShortMsg, LongMsg, and Monte tests for the given hash
   object. Transform REQ to RSP files */
static void shavstest(FILE *fin, FILE *fout, hash_object_t *h) {
  char buf[MAXLINE];
  unsigned char data[6400];
  char *line;
  int r;
  int c, i, j;
  int len;
  char *newline = "\n";
  int l;
  unsigned char hash[64];

  while (1) {
    line = fgets(buf, MAXLINE, fin);
    if (!line) {
      break;
    }
    /* check for DOS-style line endings */
    l = strlen(line);
    if (l>=2 && line[l-2] == '\r' && line[l-1] == '\n') {
      newline = "\r\n";
    }
    for (i=0; i<MAXLINE && (line[i] == ' ' || line[i] == '\t' || line[i] == '\r'); i++) {
      /* empty */
    }
    if (line[i] == '\n' || line[0] == 0 || line[0] == '#' || line[0] == '[') {
      fputs(buf, fout);
      continue;
    }
    if (strncmp(line, "Len", 3) == 0) {
      r = sscanf(line, "Len = %d", &len);
      if (r != 1 || len < 0 || len > 51200) {
	fprintf(stderr, "Error: ill-formed line:\n");
	fprintf(stderr, "%s\n", line);
	exit(1);
      }
      fputs(buf, fout);
      while (1) {
	c = fgetc(fin);
	if (c == EOF) {
	  fprintf(stderr, "Error: ill-formed file, expected '='\n");
	  exit(1);
	}
	if (c == '=') {
	  break;
	}
      }
      r = read_hex(fin, data, (len+7)/8);
      if (r < 0) {
	fprintf(stderr, "Error: ill-formed hexadecimal string\n");
	exit(1);
      }

      /* ignore rest of line */
      r = skip_line(fin);
      if (r == -1) {
	fprintf(stderr, "Error: unexpected end of file\n");
      }

      /* calculate the hash */
      h->hash_init(h->data);
      h->hash_bytes(h->data, data, len / 8);
      h->hash_finish_with_bits(h->data, len & 7, data[len/8]);
      h->hash_gethash(h->data, hash);

      if (len == 0) {
	fprintf(fout, "Msg = 00%s", newline); 
      } else {
	fprintf(fout, "Msg = ");
	fprint_hex(fout, data, (len+7)/8);
	fprintf(fout, "%s", newline);
      }
      fprintf(fout, "MD = ");
      for (i=0; i<h->hash_len; i++) {
	fprintf(fout, "%02x", hash[i]);
      }
      fprintf(fout, "%s", newline);
    } else if (strncmp(line, "Seed", 4) == 0) {
      i = 4;
      while (i<MAXLINE && line[i] != '=' && line[i] != 0) {
	i++;
      }
      if (line[i] != '=') {
	fprintf(stderr, "Error: ill-formed file, expected '='\n");
	exit(1);
      }
      r = string_hex(line+i+1, data, h->hash_len);
      if (r < 0) {
        fprintf(stderr, "Error: ill-formed hexadecimal string\n");
        exit(1);
      }

      fprintf(fout, "Seed = ");
      fprint_hex(fout, data, h->hash_len);
      fprintf(fout, "%s%s", newline, newline);
      
      /* do the monte carlo test */
      for (j=0; j<100; j++) {
	memcpy(data + h->hash_len, data, h->hash_len);
	memcpy(data + 2 * h->hash_len, data, h->hash_len);
	for (i=0; i<1000; i++) {
	  h->hash_init(h->data);
	  h->hash_bytes(h->data, data, 3 * h->hash_len);
	  h->hash_finish(h->data);
	  memmove(data, data + h->hash_len, 2 * h->hash_len);
	  h->hash_gethash(h->data, data + 2 * h->hash_len);
	}
	memcpy(data, data + 2 * h->hash_len, h->hash_len);
	fprintf(fout, "COUNT = %d%s", j, newline);
	fprintf(fout, "MD = ");
	fprint_hex(fout, data, h->hash_len);
	fprintf(fout, "%s%s", newline, newline);
      }
      return;
    }
  }
}

int main(int ac, char **av) {
  hash_object_t *h;

  if (ac != 2) {
    fprintf(stderr, "Usage: %s sha1|sha256|crc32\n", av[0]);
    return 1;
  }
  if (strcmp(av[1], "sha1") == 0) {
    h = sha1_new();
  } else if (strcmp(av[1], "sha256") == 0) {
    h = sha256_new();
  } else if (strcmp(av[1], "crc32") == 0) {
    h = crc32_new();
  } else {
    fprintf(stderr, "Usage: %s sha1|sha256|crc32\n", av[0]);
    exit(1);
  }
  if (!h) {
    fprintf(stderr, "%s: %s\n", av[0], strerror(errno));
    exit(1);
  }
  shavstest(stdin, stdout, h);
  h->hash_free(h);
  return 0;
}
