/* a reference implementation of the SHA-256 secure hash algorithm,
   defined in FIPS PUB 180-2. */

/* Copyright (C) 2000-2007 Peter Selinger. */

/* $Id: sha256.c,v 1.1 2000/12/15 19:58:38 selinger Exp $ */

/* Remarks:
   
   We only hash byte streams, not bit streams, in this implementation.

   We do not reject input stream longer than 2^64 bits. In this
   case, the "length" (for the purpose of padding) is taken mod 2^64.

   Usage: to hash a stream, let s be a pointer to a new sha256_state.
   First call sha256_init(s), then call sha256_byte(s, x) for each
   byte x to be hashed, finally call sha256_finish(s). The hash is
   then in s->H[0..7].
*/

#include <string.h>
#include <stdlib.h>

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "sha256.h"

/* ---------------------------------------------------------------------- */
/* FIPS 180-2 #3.2: Operations on Words */

static inline uint32_t rotr32(int n, uint32_t x) {
  /* this translates to a single x86 instruction */
  return (x >> n) | (x << (32-n));
}

/* FIPS 180-2 #4.1.2: SHA-256 Functions */

static inline uint32_t ch(uint32_t x, uint32_t y, uint32_t z) {
  return z ^ (x & (y ^ z));
}

static inline uint32_t maj(uint32_t x, uint32_t y, uint32_t z) {
  return (x & y) | ((x | y) & z);
}

static inline uint32_t S0(uint32_t x) {
  return rotr32(2, x) ^ rotr32(13, x) ^ rotr32(22, x);
}

static inline uint32_t S1(uint32_t x) {
  return rotr32(6, x) ^ rotr32(11, x) ^ rotr32(25, x);
}

static inline uint32_t s0(uint32_t x) {
  return rotr32(7, x) ^ rotr32(18, x) ^ (x >> 3);
}

static inline uint32_t s1(uint32_t x) {
  return rotr32(17, x) ^ rotr32(19, x) ^ (x >> 10);
}

/* ---------------------------------------------------------------------- */
/* FIPS 180-2 #4.2.2: SHA-256 Constants */

#ifdef SHA256_REFERENCE

static uint32_t K[64] = {
  0x428a2f98, 0x71374491, 0xb5c0fbcf, 0xe9b5dba5,
  0x3956c25b, 0x59f111f1, 0x923f82a4, 0xab1c5ed5,
  0xd807aa98, 0x12835b01, 0x243185be, 0x550c7dc3,
  0x72be5d74, 0x80deb1fe, 0x9bdc06a7, 0xc19bf174,
  0xe49b69c1, 0xefbe4786, 0x0fc19dc6, 0x240ca1cc,
  0x2de92c6f, 0x4a7484aa, 0x5cb0a9dc, 0x76f988da,
  0x983e5152, 0xa831c66d, 0xb00327c8, 0xbf597fc7,
  0xc6e00bf3, 0xd5a79147, 0x06ca6351, 0x14292967,
  0x27b70a85, 0x2e1b2138, 0x4d2c6dfc, 0x53380d13,
  0x650a7354, 0x766a0abb, 0x81c2c92e, 0x92722c85,
  0xa2bfe8a1, 0xa81a664b, 0xc24b8b70, 0xc76c51a3,
  0xd192e819, 0xd6990624, 0xf40e3585, 0x106aa070,
  0x19a4c116, 0x1e376c08, 0x2748774c, 0x34b0bcb5,
  0x391c0cb3, 0x4ed8aa4a, 0x5b9cca4f, 0x682e6ff3,
  0x748f82ee, 0x78a5636f, 0x84c87814, 0x8cc70208,
  0x90befffa, 0xa4506ceb, 0xbef9a3f7, 0xc67178f2,
};

#endif /* SHA256_REFERENCE */

/* FIPS 180-2 #5.3.2: SHA-256 Initial Hash Value */

static uint32_t H[8] = {
  0x6a09e667, 0xbb67ae85, 0x3c6ef372, 0xa54ff53a,
  0x510e527f, 0x9b05688c, 0x1f83d9ab, 0x5be0cd19,
};

/* ---------------------------------------------------------------------- */
/* auxiliary */

#ifdef WORDS_BIGENDIAN
# define byteorder 0
# define swap32(x) (x)
#else
# define byteorder 3
# define swap32(x) ((rotr32(24, x) & 0x00ff00ff) | (rotr32(8, x) & 0xff00ff00))
#endif /* def WORDS_BIGENDIAN */

/* ---------------------------------------------------------------------- */

#ifdef SHA256_REFERENCE

/* apply one "round" of SHA-256 hashing. Message block is supplied in
   s->W[0..15]. Implements FIPS 180-2 #6.2.2. Reference implementation. */
static void sha256_process_block(sha256_state_t *s) {
  int t;
  uint32_t a, b, c, d, e, f, g, h;
  uint32_t T1, T2;
  uint32_t W[64];

  for (t=0; t<16; t++) {
    W[t] = swap32(s->W[t]);
  }

  for (t=16; t<64; t++) {
    W[t] = s1(W[t-2]) + W[t-7] + s0(W[t-15]) + W[t-16];
  }

  a = s->H[0];
  b = s->H[1];
  c = s->H[2];
  d = s->H[3];
  e = s->H[4];
  f = s->H[5];
  g = s->H[6];
  h = s->H[7];

  for (t=0; t<64; t++) {
    T1 = h + S1(e) + ch(e, f, g) + K[t] + W[t];
    T2 = S0(a) + maj(a, b, c);
    h = g;
    g = f;
    f = e;
    e = d + T1;
    d = c;
    c = b;
    b = a;
    a = T1 + T2;
  }

  s->H[0] += a;
  s->H[1] += b;
  s->H[2] += c;
  s->H[3] += d;
  s->H[4] += e;
  s->H[5] += f;
  s->H[6] += g;
  s->H[7] += h;
}

#else /* SHA256_REFERENCE */

/* apply one "round" of SHA-256 hashing. Message block is supplied in
   s->W[0..15]. Implements FIPS 180-2 #6.2.2. Fast implementation. */
static void sha256_process_block(sha256_state_t *s) {
  uint32_t A, B, C, D, E, F, G, H;
  uint32_t W[16];
  uint32_t tmp;

  A = s->H[0];
  B = s->H[1];
  C = s->H[2];
  D = s->H[3];
  E = s->H[4];
  F = s->H[5];
  G = s->H[6];
  H = s->H[7];

  tmp = W[ 0] = swap32(s->W[ 0]);
  H = H + S1(E) + ch(E, F, G) + 0x428a2f98 + tmp;
  D += H;
  tmp = S0(A) + maj(A, B, C);
  H += tmp;

  tmp = W[ 1] = swap32(s->W[ 1]);
  G = G + S1(D) + ch(D, E, F) + 0x71374491 + tmp;
  C += G;
  tmp = S0(H) + maj(H, A, B);
  G += tmp;

  tmp = W[ 2] = swap32(s->W[ 2]);
  F = F + S1(C) + ch(C, D, E) + 0xb5c0fbcf + tmp;
  B += F;
  tmp = S0(G) + maj(G, H, A);
  F += tmp;

  tmp = W[ 3] = swap32(s->W[ 3]);
  E = E + S1(B) + ch(B, C, D) + 0xe9b5dba5 + tmp;
  A += E;
  tmp = S0(F) + maj(F, G, H);
  E += tmp;

  tmp = W[ 4] = swap32(s->W[ 4]);
  D = D + S1(A) + ch(A, B, C) + 0x3956c25b + tmp;
  H += D;
  tmp = S0(E) + maj(E, F, G);
  D += tmp;

  tmp = W[ 5] = swap32(s->W[ 5]);
  C = C + S1(H) + ch(H, A, B) + 0x59f111f1 + tmp;
  G += C;
  tmp = S0(D) + maj(D, E, F);
  C += tmp;

  tmp = W[ 6] = swap32(s->W[ 6]);
  B = B + S1(G) + ch(G, H, A) + 0x923f82a4 + tmp;
  F += B;
  tmp = S0(C) + maj(C, D, E);
  B += tmp;

  tmp = W[ 7] = swap32(s->W[ 7]);
  A = A + S1(F) + ch(F, G, H) + 0xab1c5ed5 + tmp;
  E += A;
  tmp = S0(B) + maj(B, C, D);
  A += tmp;

  tmp = W[ 8] = swap32(s->W[ 8]);
  H = H + S1(E) + ch(E, F, G) + 0xd807aa98 + tmp;
  D += H;
  tmp = S0(A) + maj(A, B, C);
  H += tmp;

  tmp = W[ 9] = swap32(s->W[ 9]);
  G = G + S1(D) + ch(D, E, F) + 0x12835b01 + tmp;
  C += G;
  tmp = S0(H) + maj(H, A, B);
  G += tmp;

  tmp = W[10] = swap32(s->W[10]);
  F = F + S1(C) + ch(C, D, E) + 0x243185be + tmp;
  B += F;
  tmp = S0(G) + maj(G, H, A);
  F += tmp;

  tmp = W[11] = swap32(s->W[11]);
  E = E + S1(B) + ch(B, C, D) + 0x550c7dc3 + tmp;
  A += E;
  tmp = S0(F) + maj(F, G, H);
  E += tmp;

  tmp = W[12] = swap32(s->W[12]);
  D = D + S1(A) + ch(A, B, C) + 0x72be5d74 + tmp;
  H += D;
  tmp = S0(E) + maj(E, F, G);
  D += tmp;

  tmp = W[13] = swap32(s->W[13]);
  C = C + S1(H) + ch(H, A, B) + 0x80deb1fe + tmp;
  G += C;
  tmp = S0(D) + maj(D, E, F);
  C += tmp;

  tmp = W[14] = swap32(s->W[14]);
  B = B + S1(G) + ch(G, H, A) + 0x9bdc06a7 + tmp;
  F += B;
  tmp = S0(C) + maj(C, D, E);
  B += tmp;

  tmp = W[15] = swap32(s->W[15]);
  A = A + S1(F) + ch(F, G, H) + 0xc19bf174 + tmp;
  E += A;
  tmp = S0(B) + maj(B, C, D);
  A += tmp;

  tmp = W[ 0] += s1(W[14]) + W[ 9] + s0(W[ 1]);
  H = H + S1(E) + ch(E, F, G) + 0xe49b69c1 + tmp;
  D += H;
  tmp = S0(A) + maj(A, B, C);
  H += tmp;

  tmp = W[ 1] += s1(W[15]) + W[10] + s0(W[ 2]);
  G = G + S1(D) + ch(D, E, F) + 0xefbe4786 + tmp;
  C += G;
  tmp = S0(H) + maj(H, A, B);
  G += tmp;

  tmp = W[ 2] += s1(W[ 0]) + W[11] + s0(W[ 3]);
  F = F + S1(C) + ch(C, D, E) + 0x0fc19dc6 + tmp;
  B += F;
  tmp = S0(G) + maj(G, H, A);
  F += tmp;

  tmp = W[ 3] += s1(W[ 1]) + W[12] + s0(W[ 4]);
  E = E + S1(B) + ch(B, C, D) + 0x240ca1cc + tmp;
  A += E;
  tmp = S0(F) + maj(F, G, H);
  E += tmp;

  tmp = W[ 4] += s1(W[ 2]) + W[13] + s0(W[ 5]);
  D = D + S1(A) + ch(A, B, C) + 0x2de92c6f + tmp;
  H += D;
  tmp = S0(E) + maj(E, F, G);
  D += tmp;

  tmp = W[ 5] += s1(W[ 3]) + W[14] + s0(W[ 6]);
  C = C + S1(H) + ch(H, A, B) + 0x4a7484aa + tmp;
  G += C;
  tmp = S0(D) + maj(D, E, F);
  C += tmp;

  tmp = W[ 6] += s1(W[ 4]) + W[15] + s0(W[ 7]);
  B = B + S1(G) + ch(G, H, A) + 0x5cb0a9dc + tmp;
  F += B;
  tmp = S0(C) + maj(C, D, E);
  B += tmp;

  tmp = W[ 7] += s1(W[ 5]) + W[ 0] + s0(W[ 8]);
  A = A + S1(F) + ch(F, G, H) + 0x76f988da + tmp;
  E += A;
  tmp = S0(B) + maj(B, C, D);
  A += tmp;

  tmp = W[ 8] += s1(W[ 6]) + W[ 1] + s0(W[ 9]);
  H = H + S1(E) + ch(E, F, G) + 0x983e5152 + tmp;
  D += H;
  tmp = S0(A) + maj(A, B, C);
  H += tmp;

  tmp = W[ 9] += s1(W[ 7]) + W[ 2] + s0(W[10]);
  G = G + S1(D) + ch(D, E, F) + 0xa831c66d + tmp;
  C += G;
  tmp = S0(H) + maj(H, A, B);
  G += tmp;

  tmp = W[10] += s1(W[ 8]) + W[ 3] + s0(W[11]);
  F = F + S1(C) + ch(C, D, E) + 0xb00327c8 + tmp;
  B += F;
  tmp = S0(G) + maj(G, H, A);
  F += tmp;

  tmp = W[11] += s1(W[ 9]) + W[ 4] + s0(W[12]);
  E = E + S1(B) + ch(B, C, D) + 0xbf597fc7 + tmp;
  A += E;
  tmp = S0(F) + maj(F, G, H);
  E += tmp;

  tmp = W[12] += s1(W[10]) + W[ 5] + s0(W[13]);
  D = D + S1(A) + ch(A, B, C) + 0xc6e00bf3 + tmp;
  H += D;
  tmp = S0(E) + maj(E, F, G);
  D += tmp;

  tmp = W[13] += s1(W[11]) + W[ 6] + s0(W[14]);
  C = C + S1(H) + ch(H, A, B) + 0xd5a79147 + tmp;
  G += C;
  tmp = S0(D) + maj(D, E, F);
  C += tmp;

  tmp = W[14] += s1(W[12]) + W[ 7] + s0(W[15]);
  B = B + S1(G) + ch(G, H, A) + 0x06ca6351 + tmp;
  F += B;
  tmp = S0(C) + maj(C, D, E);
  B += tmp;

  tmp = W[15] += s1(W[13]) + W[ 8] + s0(W[ 0]);
  A = A + S1(F) + ch(F, G, H) + 0x14292967 + tmp;
  E += A;
  tmp = S0(B) + maj(B, C, D);
  A += tmp;

  tmp = W[ 0] += s1(W[14]) + W[ 9] + s0(W[ 1]);
  H = H + S1(E) + ch(E, F, G) + 0x27b70a85 + tmp;
  D += H;
  tmp = S0(A) + maj(A, B, C);
  H += tmp;

  tmp = W[ 1] += s1(W[15]) + W[10] + s0(W[ 2]);
  G = G + S1(D) + ch(D, E, F) + 0x2e1b2138 + tmp;
  C += G;
  tmp = S0(H) + maj(H, A, B);
  G += tmp;

  tmp = W[ 2] += s1(W[ 0]) + W[11] + s0(W[ 3]);
  F = F + S1(C) + ch(C, D, E) + 0x4d2c6dfc + tmp;
  B += F;
  tmp = S0(G) + maj(G, H, A);
  F += tmp;

  tmp = W[ 3] += s1(W[ 1]) + W[12] + s0(W[ 4]);
  E = E + S1(B) + ch(B, C, D) + 0x53380d13 + tmp;
  A += E;
  tmp = S0(F) + maj(F, G, H);
  E += tmp;

  tmp = W[ 4] += s1(W[ 2]) + W[13] + s0(W[ 5]);
  D = D + S1(A) + ch(A, B, C) + 0x650a7354 + tmp;
  H += D;
  tmp = S0(E) + maj(E, F, G);
  D += tmp;

  tmp = W[ 5] += s1(W[ 3]) + W[14] + s0(W[ 6]);
  C = C + S1(H) + ch(H, A, B) + 0x766a0abb + tmp;
  G += C;
  tmp = S0(D) + maj(D, E, F);
  C += tmp;

  tmp = W[ 6] += s1(W[ 4]) + W[15] + s0(W[ 7]);
  B = B + S1(G) + ch(G, H, A) + 0x81c2c92e + tmp;
  F += B;
  tmp = S0(C) + maj(C, D, E);
  B += tmp;

  tmp = W[ 7] += s1(W[ 5]) + W[ 0] + s0(W[ 8]);
  A = A + S1(F) + ch(F, G, H) + 0x92722c85 + tmp;
  E += A;
  tmp = S0(B) + maj(B, C, D);
  A += tmp;

  tmp = W[ 8] += s1(W[ 6]) + W[ 1] + s0(W[ 9]);
  H = H + S1(E) + ch(E, F, G) + 0xa2bfe8a1 + tmp;
  D += H;
  tmp = S0(A) + maj(A, B, C);
  H += tmp;

  tmp = W[ 9] += s1(W[ 7]) + W[ 2] + s0(W[10]);
  G = G + S1(D) + ch(D, E, F) + 0xa81a664b + tmp;
  C += G;
  tmp = S0(H) + maj(H, A, B);
  G += tmp;

  tmp = W[10] += s1(W[ 8]) + W[ 3] + s0(W[11]);
  F = F + S1(C) + ch(C, D, E) + 0xc24b8b70 + tmp;
  B += F;
  tmp = S0(G) + maj(G, H, A);
  F += tmp;

  tmp = W[11] += s1(W[ 9]) + W[ 4] + s0(W[12]);
  E = E + S1(B) + ch(B, C, D) + 0xc76c51a3 + tmp;
  A += E;
  tmp = S0(F) + maj(F, G, H);
  E += tmp;

  tmp = W[12] += s1(W[10]) + W[ 5] + s0(W[13]);
  D = D + S1(A) + ch(A, B, C) + 0xd192e819 + tmp;
  H += D;
  tmp = S0(E) + maj(E, F, G);
  D += tmp;

  tmp = W[13] += s1(W[11]) + W[ 6] + s0(W[14]);
  C = C + S1(H) + ch(H, A, B) + 0xd6990624 + tmp;
  G += C;
  tmp = S0(D) + maj(D, E, F);
  C += tmp;

  tmp = W[14] += s1(W[12]) + W[ 7] + s0(W[15]);
  B = B + S1(G) + ch(G, H, A) + 0xf40e3585 + tmp;
  F += B;
  tmp = S0(C) + maj(C, D, E);
  B += tmp;

  tmp = W[15] += s1(W[13]) + W[ 8] + s0(W[ 0]);
  A = A + S1(F) + ch(F, G, H) + 0x106aa070 + tmp;
  E += A;
  tmp = S0(B) + maj(B, C, D);
  A += tmp;

  tmp = W[ 0] += s1(W[14]) + W[ 9] + s0(W[ 1]);
  H = H + S1(E) + ch(E, F, G) + 0x19a4c116 + tmp;
  D += H;
  tmp = S0(A) + maj(A, B, C);
  H += tmp;

  tmp = W[ 1] += s1(W[15]) + W[10] + s0(W[ 2]);
  G = G + S1(D) + ch(D, E, F) + 0x1e376c08 + tmp;
  C += G;
  tmp = S0(H) + maj(H, A, B);
  G += tmp;

  tmp = W[ 2] += s1(W[ 0]) + W[11] + s0(W[ 3]);
  F = F + S1(C) + ch(C, D, E) + 0x2748774c + tmp;
  B += F;
  tmp = S0(G) + maj(G, H, A);
  F += tmp;

  tmp = W[ 3] += s1(W[ 1]) + W[12] + s0(W[ 4]);
  E = E + S1(B) + ch(B, C, D) + 0x34b0bcb5 + tmp;
  A += E;
  tmp = S0(F) + maj(F, G, H);
  E += tmp;

  tmp = W[ 4] += s1(W[ 2]) + W[13] + s0(W[ 5]);
  D = D + S1(A) + ch(A, B, C) + 0x391c0cb3 + tmp;
  H += D;
  tmp = S0(E) + maj(E, F, G);
  D += tmp;

  tmp = W[ 5] += s1(W[ 3]) + W[14] + s0(W[ 6]);
  C = C + S1(H) + ch(H, A, B) + 0x4ed8aa4a + tmp;
  G += C;
  tmp = S0(D) + maj(D, E, F);
  C += tmp;

  tmp = W[ 6] += s1(W[ 4]) + W[15] + s0(W[ 7]);
  B = B + S1(G) + ch(G, H, A) + 0x5b9cca4f + tmp;
  F += B;
  tmp = S0(C) + maj(C, D, E);
  B += tmp;

  tmp = W[ 7] += s1(W[ 5]) + W[ 0] + s0(W[ 8]);
  A = A + S1(F) + ch(F, G, H) + 0x682e6ff3 + tmp;
  E += A;
  tmp = S0(B) + maj(B, C, D);
  A += tmp;

  tmp = W[ 8] += s1(W[ 6]) + W[ 1] + s0(W[ 9]);
  H = H + S1(E) + ch(E, F, G) + 0x748f82ee + tmp;
  D += H;
  tmp = S0(A) + maj(A, B, C);
  H += tmp;

  tmp = W[ 9] += s1(W[ 7]) + W[ 2] + s0(W[10]);
  G = G + S1(D) + ch(D, E, F) + 0x78a5636f + tmp;
  C += G;
  tmp = S0(H) + maj(H, A, B);
  G += tmp;

  tmp = W[10] += s1(W[ 8]) + W[ 3] + s0(W[11]);
  F = F + S1(C) + ch(C, D, E) + 0x84c87814 + tmp;
  B += F;
  tmp = S0(G) + maj(G, H, A);
  F += tmp;

  tmp = W[11] += s1(W[ 9]) + W[ 4] + s0(W[12]);
  E = E + S1(B) + ch(B, C, D) + 0x8cc70208 + tmp;
  A += E;
  tmp = S0(F) + maj(F, G, H);
  E += tmp;

  tmp = W[12] += s1(W[10]) + W[ 5] + s0(W[13]);
  D = D + S1(A) + ch(A, B, C) + 0x90befffa + tmp;
  H += D;
  tmp = S0(E) + maj(E, F, G);
  D += tmp;

  tmp = W[13] += s1(W[11]) + W[ 6] + s0(W[14]);
  C = C + S1(H) + ch(H, A, B) + 0xa4506ceb + tmp;
  G += C;
  tmp = S0(D) + maj(D, E, F);
  C += tmp;

  tmp = W[14] += s1(W[12]) + W[ 7] + s0(W[15]);
  B = B + S1(G) + ch(G, H, A) + 0xbef9a3f7 + tmp;
  F += B;
  tmp = S0(C) + maj(C, D, E);
  B += tmp;

  tmp = W[15] += s1(W[13]) + W[ 8] + s0(W[ 0]);
  A = A + S1(F) + ch(F, G, H) + 0xc67178f2 + tmp;
  E += A;
  tmp = S0(B) + maj(B, C, D);
  A += tmp;

  s->H[0] += A;
  s->H[1] += B;
  s->H[2] += C;
  s->H[3] += D;
  s->H[4] += E;
  s->H[5] += F;
  s->H[6] += G;
  s->H[7] += H;
}

#endif /* SHA256_REFERENCE */

/* initialize SHA-256 state */
void sha256_init(sha256_state_t *s) {
  int i;
  s->count_lo = 0;
  s->count_hi = 0;
  s->j = 0;
  for (i=0; i<8; i++) {
    s->H[i] = H[i];
  }
}

/* hash the next byte */
inline void sha256_byte(sha256_state_t *s, int x) {
  unsigned char *p = (unsigned char *)&s->W;  /* access s->W as byte array */

  p[s->j] = x;
  s->j++;
  s->count_lo += 8;
  if (s->count_lo == 0) {
    s->count_hi += 1;
  }
  if (s->j == 64) {
    sha256_process_block(s);
    s->j = 0;
  }
}

#ifdef SHA256_REFERENCE

/* hash multiple bytes - reference implementation */
void sha256_bytes(sha256_state_t *s, unsigned char *p, int n) {
  int i;

  for (i=0; i<n; i++) {
    sha256_byte(s, p[i]);
  }
}

#else /* SHA256_REFERENCE */

/* hash multiple bytes - fast implementation */
void sha256_bytes(sha256_state_t *s, unsigned char *p, int n) {

  /* do the general case, one byte at a time */
  while (n>0 && s->j != 0) {
    sha256_byte(s, *p);
    p++;
    n--;
  }

  /* do the typical case, one block at a time */
  while (n>=64) {
    memcpy(s->W, p, 64);
    sha256_process_block(s);
    p+=64;
    n-=64;
    s->count_lo += 512;
    if (s->count_lo == 0) {
      s->count_hi += 1;
    }
  }

  /* finish with general case, one byte at a time */
  while (n>0) {
    sha256_byte(s, *p);
    p++;
    n--;
  }
}

#endif /* SHA256_REFERENCE */

/* Add 0 to 7 bits to the end of the stream, then finish up hashing
   and return final hash. The bits are taken as the n most significant
   bits of the byte buf */
inline void sha256_finish_with_bits(sha256_state_t *s, int n, unsigned char buf) {
  uint32_t count_lo = s->count_lo + n;
  uint32_t count_hi = s->count_hi;

  /* first apply padding */
  buf &= 0xff << (8-n);
  buf |= 0x80 >> n;

  sha256_byte(s, buf);
  while (s->j != 56) {
    sha256_byte(s, 0x00);
  }
  s->W[14] = swap32(count_hi);
  s->W[15] = swap32(count_lo);

  /* then process last block */
  sha256_process_block(s);
}

/* finish up hashing and return final hash */
void sha256_finish(sha256_state_t *s) {
  sha256_finish_with_bits(s, 0, 0);
}

void sha256_gethash(sha256_state_t *s, unsigned char *p) {
  int i;
  unsigned char *q = (unsigned char *)(s->H);

  for (i=0; i<32; i++) {
    p[i] = q[i ^ byteorder];
  }
}

static void sha256_free(hash_object_t *s) {
  if (s != NULL) {
    free(s->data);
    s->data = NULL;
  }
  free(s);
}

/* return NULL on error with errno set, else an initialized sha256 hash
   object.  */
hash_object_t *sha256_new(void) {
  hash_object_t *self;
  sha256_state_t *data;

  self = (hash_object_t *)malloc(sizeof(hash_object_t));
  if (!self) {
    return NULL;
  }
  data = (sha256_state_t *)malloc(sizeof(sha256_state_t));
  if (!data) {
    free(self);
    return NULL;
  }
  sha256_init(data);

  self->data = data;
  self->hash_len = 32;
  self->hash_init = (void (*)(void *)) sha256_init;
  self->hash_byte = (void (*)(void *, int)) sha256_byte;
  self->hash_bytes = (void (*)(void *, unsigned char *, int)) sha256_bytes;
  self->hash_finish = (void (*)(void *)) sha256_finish;
  self->hash_finish_with_bits = (void (*)(void *, int, unsigned char)) sha256_finish_with_bits;
  self->hash_gethash = (void (*)(void *, unsigned char *)) sha256_gethash;
  self->hash_free = sha256_free;

  return self;
};
