/* $Id: traverse.c 39 2007-01-17 06:26:22Z selinger $ */ 
/* directory-traversing user interface for sha1/sha256 */

#include <stdio.h>
#include <sys/stat.h>
#include <dirent.h>
#include <limits.h>
#include <getopt.h>
#include <string.h>
#include <stdlib.h>
#include <errno.h>

#ifdef HAVE_CONFIG_H
# include <config.h>
#endif

#include "hash.h"

#define BUFSIZE 409600

/* ---------------------------------------------------------------------- */
/* global variables set by command line options */

static int recursive = 0;
static int quiet = 0;

/* ---------------------------------------------------------------------- */
/* functions to traverse files and directories */

/* prototypes */

static void do_file(FILE *file, char *filename, hash_object_t *h);
static void do_filename(char *filename, hash_object_t *h);
static void do_directory(char *dirname, hash_object_t *h);

/* process a file that has already been opened; allow filename=0 in
   cases such as stdin. */

static void do_file(FILE *file, char *filename, hash_object_t *h) {
  int i, r;
  static unsigned char buf[BUFSIZE];
  unsigned char hash[64];

  h->hash_init(h->data);
  while (1) {
    r = fread(buf, 1, BUFSIZE, file);
    h->hash_bytes(h->data, buf, r);
    if (r < BUFSIZE) {
      break;
    }
  }
  h->hash_finish(h->data);
  h->hash_gethash(h->data, hash);

  for (i=0; i<h->hash_len; i++) {
    printf("%02x", hash[i]);
  }
  if (filename)
    printf(" %s", filename);
  printf("\n");
  fflush(stdout);
}

/* process a file or directory by name */

static void do_filename(char *filename, hash_object_t *h) {
  FILE *file;
  struct stat filestat;
  int st;

  /* get info about file */
  st = stat(filename, &filestat);
  if (st) {
    fprintf(stderr, "%s: %s\n", filename, strerror(errno));
    return;
  }

  /* is it a directory? */
  if (S_ISDIR(filestat.st_mode)) {
    if (recursive) {
      do_directory(filename, h);
    } else if (!quiet) {
      fprintf(stderr, "%s: is a directory\n", filename);
    }
    return;
  }

  /* else we assume it is an ordinary file */

  file = fopen(filename, "r");
  if (file==NULL) {
    fprintf(stderr, "%s: %s\n", filename, strerror(errno));
    return;
  } 
  do_file(file, filename, h);
  fclose(file);
}

/* recursively traverse all files in the given directory */

static void do_directory(char *dirname, hash_object_t *h) {
  DIR *dir;
  struct dirent *dirent;
  
  dir = opendir(dirname);
  if (dir==NULL) {
    fprintf(stderr, "%s: %s\n", dirname, strerror(errno));
    return;
  }

  while ((dirent = readdir(dir)) != NULL) {
    if (strcmp(dirent->d_name, "..")!=0 && strcmp(dirent->d_name, ".")!=0) {
      int len = strlen(dirname)+strlen(dirent->d_name)+2;
      char filename[len];
      strcpy(filename, dirname);
      strcat(filename, "/");
      strcat(filename, dirent->d_name);
      do_filename(filename, h);
    }
  }

  closedir(dir);
}

/* ---------------------------------------------------------------------- */
/* usage, version, license */

void usage(FILE *fout, char *name, char *shortdesc) {
  fprintf(fout, "%s %s. %s\n", name, VERSION, shortdesc);
  fprintf(fout, "Usage: %s [options] [file...]\n", name);
  fprintf(fout, 
"Options: -h, --help         print this help message and exit\n"
"         -v, --version      print version info and exit\n"
"         -l, --license      print license info and exit\n"
"         -r, --recursive    traverse subdirectories recursively\n"
"         -q, --quiet        suppress warnings\n"
);
}

void version(FILE *fout, char *name, char *shortdesc) {
  fprintf(fout, "%s %s. Copyright (C) 2000-2007 Peter Selinger\n", name, VERSION);
}

void license(FILE *fout, char *name, char *shortdesc) {
  fprintf(fout, "%s %s. Copyright (C) 2000-2007 Peter Selinger\n", name, VERSION);
  fprintf(fout, "\n"
  "This program is free software; you can redistribute it and/or modify\n"
  "it under the terms of the GNU General Public License as published by\n"
  "the Free Software Foundation; either version 2 of the License, or\n"
  "(at your option) any later version.\n"
  "\n"
  "This program is distributed in the hope that it will be useful,\n"
  "but WITHOUT ANY WARRANTY; without even the implied warranty of\n"
  "MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the\n"
  "GNU General Public License for more details.\n"
  "\n"
  "You should have received a copy of the GNU General Public License\n"
  "along with this program; if not, write to the Free Software\n"
  "Foundation, Inc., 59 Temple Place, Suite 330, Boston, MA 02111-1307, USA.\n"
	  );
}

/* ---------------------------------------------------------------------- */
/* main */

static struct option longopts[] = {
  {"help", 0, 0, 'h'},
  {"version", 0, 0, 'v'},
  {"license", 0, 0, 'l'},
  {"recursive", 0, 0, 'r'},
  {"quiet", 0, 0, 'q'},
  {0, 0, 0, 0}
};

static char *shortopts = "hvlrq";

/* this is called directly from main(). */
int traverse(int ac, char *av[], hash_object_t *h, char *name, char *shortdesc) {
  int c;

  /* read options */
  while ((c = getopt_long(ac, av, shortopts, longopts, NULL)) != -1) {
    switch (c) {
    case 'h':
      usage(stdout, name, shortdesc);
      exit(0);
    case 'v':
      version(stdout, name, shortdesc);
      exit(0);
    case 'l':
      license(stdout, name, shortdesc);
      exit(0);
    case 'r':
      recursive=1;
      break;
    case 'q':
      quiet=1;
      break;
    default:
      fprintf(stderr, "%s: illegal option %c. \n", name, c);
      /* fall through */
    case '?':
      fprintf(stderr, "Try --help for more information.\n");
      exit(1);
    }
  }

  /* if no filenames follow, act as a filter */
  if (optind == ac) {
    do_file(stdin, 0, h);
  } else {
    /* otherwise, act on file arguments */
    while (optind < ac) {
      do_filename(av[optind], h);
      optind++;
    }
  }
  h->hash_free(h);
  return 0;
}
